
#define PROGNAM "PolyEM multivar"
#define PNAM "multivar"

#include "polyem.h"

int chfact(double *a, double *b, int n, double *ldet);
void chsolve(double *a, double *y, double *z, int n);
void chinv(double *a, double *b, int n);

/*
 * program to test basic em equations : case of fixed + random + error 
 *
 * This May 1991 version uses the full G-inverse, instead of the G for 
 * observed individuals. It is adapted for a large pedigree, and takes as 
 * input the components of a sparse G-inverse, computed for all 
 * individuals, and also a set of eigenvalues of the G (hence G-inverse)
 * for observed individuals. 
 *
 * Evgen will now instead contain the eigenvalues of the INVERSE of the 
 * G for the observed individuals 
 *
 * modified (especially with regard to fixed effects) for greater 
 * flexibility and compatibility ; May 1992 
 *
 * Modified for correct behaviour with looped pedigrees; June 1997 (SCH) 
 *
 * Modified to use lower triangle for all half-stored matrices; 
 *                                                         July 1997 (SCH) 
 *
 * modified, Aug 1997, by MAJ, EAT and SCH for Version 2.0 revisions:   
 * use of parameter file for user-friendly specification of modelling 
 * parameters;
 * use of dynamic storage for all major arrays, with memory allocated just 
 * prior to use, and freed as soon as no longer needed;
 */

/* Access element b[i][j] from half-stored lower triangle of symmetric */
/* matrix size n Used by mmm(), chfact(), chsolve() and chinv(). */

#define BB(b,i,j) ((b)[i*(i+1)/2+j])
#define BB1(b,i,j) ((i<j)?BB(b,j,i):BB(b,i,j))

void mmm(double *m1, double *m2, double *m3, double *m_out, int nt)
{
    int i, j, k;
    double z;

    for (i = 0; i < nt * (nt + 1) / 2; i++)
	m_out[i] = 0.0;
    for (i = 0; i < nt; i++)
    {
	for (j = 0; j < nt; j++)
	{
	    for (z = 0.0, k = 0; k < nt; k++)
		z += BB1(m1, i, k) * BB1(m2, k, j);
	    for (k = i; k < nt; k++)
		BB(m_out, k, i) += z * BB1(m3, j, k);
	}
    }
}

void chk_sgfit(Term * ginv, int nt, double **b, double *ast, 
                                    int NTraits, int *SigModel)
{
    int i, j, k;
    double tmp;

    k = 0;
    for (i = 0; i < NTraits; ++i)
	for (j = 0; j <= i; ++j)
	{
            if (SigModel[k])
            {
                tmp = quarz(b[i], b[j], ginv, nt);
                SigSq[k] = tmp / np + ast[k];
                printf(" check sig%d%d;  trace =%11.3f,  quad =%11.3f,  "
                       "net =%8.3f\n", i + 1 , j + 1, ast[k] * np, 
                                                           tmp, SigSq[k]);
            }
            else
                SigSq[k] = 0.0;
	    ++k;
	}
    printf("\n");
}

void chk_tufit(double **e, double *tst, int NTraits, int *TauModel)
{
    int i, j, k;
    double tmp;

    k = 0;

    for (i = 0; i < NTraits; ++i)
	for (j = 0; j <= i; ++j)
	{
            if (TauModel[k])
            {
               tmp = innerprod(e[i], e[j], np); 
               TauSq[k] = tmp / np + tst[k];
               printf(" check tau%d%d;  trace =%11.3f,  quad =%11.3f,  "
                       "net =%8.3f\n", i + 1, j + 1, tst[k] * np, 
                                                          tmp, TauSq[k]);
            }
            else
               TauSq[k] = 0.0;
	    ++k;
	}
    printf("\n\n");
}

void sig_fit(Term * ginv, int nt, double **b, double *ast, 
                                              int NTraits, int *SigModel)
{
    int i, j, k;

    k = 0;
    for (i = 0; i < NTraits; ++i)
	for (j = 0; j <= i; ++j)
	{
	    SigSq[k] = SigModel[k] ? 
                        (quarz(b[i], b[j], ginv, nt)) / np + ast[k] : 0.0;
	    ++k;
	}
}

void tau_fit(double **e, double *tst, int NTraits, int *TauModel)
{
    int i, j, k;

    k = 0;

    for (i = 0; i < NTraits; ++i)
	for (j = 0; j <= i; ++j)
	{
	    TauSq[k] = TauModel[k] ? 
                        (innerprod(e[i], e[j], np)) / np + tst[k] : 0.0;
	    ++k;
	}
}

double ss(double ev[], double *ast, double *tst, int nt)
{
    int i, j, dd = nt * (nt + 1) / 2;
    double cc, ldet;
    double *inv1, *inv;
    double *ss;

    inv = malloc(dd * 3 * sizeof(double));
    alloc_chk(WHERE, inv);
    inv1 = inv + dd;
    ss = inv1 + dd;

    /* find the 'lamda * A + T' matrix with the first (smallest) */
    /* eigenvlaue */

    for (j = 0; j < dd; ++j)
	ss[j] = SigSq[j] / ev[1] + TauSq[j];

    if (chfact(ss, ss, nt, &ldet))
        give_up_where(WHERE, ERR_mtx, "ss");
    cc = ldet;
    chinv(ss, inv1, nt);
    for (j = 0; j < dd; j++)
	inv[j] = inv1[j];

    for (i = 2; i < np + 1; ++i)
    {
	/* if the evgenvalue changes, recompute the 'lamda * A +T' 
         * matrix, 
	 * find its det, and inverse matrix */

	if (ev[i] != ev[i - 1])
	{
	    for (j = 0; j < dd; ++j)
		ss[j] = SigSq[j] / ev[i] + TauSq[j];
	    if (chfact(ss, ss, nt, &ldet))
                give_up_where(WHERE, ERR_mtx, "ss");
	    chinv(ss, inv1, nt);
	}
	cc += ldet;
	for (j = 0; j < dd; ++j)
	    inv[j] += inv1[j];
    }

    for (j = 0; j < dd; ++j)
	inv[j] = inv[j] / np;

    if (ast && tst)
    {
        mmm(SigSq, inv, TauSq, ast, nt);
        mmm(TauSq, inv, TauSq, tst, nt);
        for (j = 0; j < dd; ++j)
	    tst[j] = TauSq[j] - tst[j];
    }

    free((void *) inv);

    return (cc);
}

/* Compute Cholesky factorization of symmetric matrix. Inputs:  a - Lower
 * triangle of matrix n - Size of matrix 
 *
 * Outputs: b - Cholesky factor of a ldet - log determinant of a 
 *
 * Returns 1 if a is not positive definite, otherwise returns 0 
 *
 * Note that b can point to the same location as a, otherwise a is 
 * unmodified */

int chfact(double *a, double *b, int n, double *ldet)
{
    int i, j, k, kk;
    double zz, piv;

    kk = n * (n + 1) / 2;
    *ldet = 0.0;
    for (i = n - 1; i >= 0; i--)
    {
	zz = a[--kk];
	for (k = i + 1; k < n; k++)
	    zz -= BB(b, k, i) * BB(b, k, i);
	if (zz <= 0.0)
	    return -1;
	b[kk] = piv = sqrt(zz);
	*ldet += log(zz);
	for (j = i - 1; j >= 0; j--)
	{
	    zz = a[--kk];
	    for (k = i + 1; k < n; k++)
		zz -= BB(b, k, i) * BB(b, k, j);
	    b[kk] = zz / piv;
	}
    }
    return 0;
}

/* Solve A*y=z 
 *
 * Inputs:  a - Cholesky factor of A computed by chfact() 
 *          y - Right hand side
 *          vector n - Size of A and y 
 *
 * Outputs: z - Solution of A*y=z 
 *
 * a is not modified.  The same is true for y unless y and z point
 * to the same location, which is OK 
 */

void chsolve(double *a, double *y, double *z, int n)
{
    int i, j, kk;
    double zz;

    for (i = n - 1; i >= 0; i--)
    {
	zz = y[i];
	for (j = i + 1; j < n; j++)
	    zz -= BB(a, j, i) * z[j];
	z[i] = zz / BB(a, i, i);
    }
    kk = 0;
    for (i = 0; i < n; i++)
    {
	zz = z[i];
	for (j = 0; j < i; j++)
	    zz -= a[kk++] * z[j];
	z[i] = zz / a[kk++];
    }
}

/*
 * Compute inverse of symmetrix pd matrix A 
 *
 * Inputs:  a - Cholesky factor of A computed by chfact() n - Size of A 
 *
 * Outputs: b - (half stored) inverse of A 
 *
 * Note that b can point to the same location as a, otherwise a is 
 * unmodified 
 */

void chinv(double *a, double *b, int n)
{
    int i, j, k, kk;
    double zz;

    kk = n * (n + 1) / 2;
    for (i = n - 1; i >= 0; i--)
    {
	zz = BB(a, i, i);
	b[--kk] = 1.0 / zz;
	for (j = i - 1; j >= 0; j--)
	{
	    zz = 0.0;
	    for (k = i; k > j; k--)
		zz -= BB(a, k, j) * BB(b, i, k);
	    b[--kk] = zz / BB(a, j, j);
	}
    }
    kk = n * (n + 1) / 2;
    for (i = n - 1; i >= 0; i--)
    {
	for (j = i; j >= 0; j--)
	{
	    zz = 0.0;
	    for (k = j; k >= 0; k--)
		zz += BB(b, i, k) * BB(b, j, k);
	    b[--kk] = zz;
	}
    }
}

/* Note within this routine "xx" are the residuals from any fixed effects;
 * i.e. "y" and NOT original data x 
 *
 * Re-written to allow for looped pedigrees; June 1997 (SCH) 
 *
 * IMPORTANT: Assumes that ginv contains the *lower left* non-zero terms 
 * of the full G-inverse.  The henderson() routine has been modified to 
 * do this (originally it produced the *upper right* terms).  This, of 
 * course, only affects the order in which the terms are stored. Having 
 * the lower left terms means that itermo() can efficiently work from the 
 * bottom of the pedigree upwards, which improves initial convergence if 
 * data is mainly present at the bottom of the pedigree. 
 */

void itermo(double **xx, double **b, int NTraits, Term * ginv, int nt)
{
    int i, j, k, k1, n, dd, l, it;
    double z, z1, *cc, *cfact, *d, *zz, *zz1, 
                  *sigch, *tauch, *siginv, *tauinv, **tt;

    dd = NTraits * (NTraits + 1) / 2;

    cc = malloc(sizeof(double) * 
                             (6 * dd + 3 * NTraits + NTraits * pedsize));
    alloc_chk(WHERE, cc);
    tt = malloc(sizeof(void *) * NTraits);
    alloc_chk(WHERE, tt);

    cfact = cc + dd;
    sigch = cfact + dd;
    tauch = sigch + dd;
    siginv = tauch + dd;
    tauinv = siginv + dd;
    d = tauinv + dd;
    zz = d + NTraits;
    zz1 = zz + NTraits;
    tt[0] = zz1 + NTraits;
    for (i = 1; i < NTraits; i++)
	tt[i] = tt[i - 1] + pedsize;

    /* Factor covariance matrices, and calculate inverses */
    /* (We need both the factors and the inverses later */
    if (chfact(SigSq, sigch, NTraits, &z))
        give_up_where(WHERE, ERR_mxg, "itermo");
    if (chfact(TauSq, tauch, NTraits, &z))
        give_up_where(WHERE, ERR_mxe, "itermo");
    chinv(sigch, siginv, NTraits);
    chinv(tauch, tauinv, NTraits);

    for (it = 0; it < ItMo; it++)
    {
	/* Deal with observations on everyone for all traits */
	for (i = 0; i < pedsize; i++)
	{
	    if (j = qdex[i])
	    {
		for (l = 0; l < NTraits; l++)
		    zz1[l] = xx[l][j];
		chsolve(tauch, zz1, zz, NTraits);
		for (l = 0; l < NTraits; l++)
		    tt[l][i] = zz[l];
	    } else
		for (l = 0; l < NTraits; l++)
		    tt[l][i] = 0.0;
	}

	for (i = 0; i < NTraits; i++)
	    d[i] = 0.0;
	n = 0;
	k = nt;				/* Start with last term of ginv */
					/* (bottom of pedigree) */
	do
	{
	    /* Pick up diagonal element of ginv */
	    i = ginv[k].r1;
	    z = ginv[k--].value;
	    for (j = 0; j < dd; j++)
		cc[j] = siginv[j] * z;
	    if (qdex[i])
		for (j = 0; j < dd; j++)
		    cc[j] += tauinv[j];
	    chfact(cc, cfact, NTraits, &z1);
	    k1 = k;

	    /* Go through and deal with off-diagonal elements (j<i) */
	    while (k > 0 && ginv[k].r1 == i)
	    {
		j = ginv[k].r2;
		z = ginv[k--].value;
		for (l = 0; l < NTraits; l++)
		    zz1[l] = b[l][j];
		chsolve(sigch, zz1, zz, NTraits);
		for (l = 0; l < NTraits; l++)
		    tt[l][i] -= z * zz[l];
	    }
	    /* Compute zz=inv(cc)*tt[l] */
	    for (l = 0; l < NTraits; l++)
		zz1[l] = tt[l][i];
	    chsolve(cfact, zz1, zz, NTraits);

	    /* zz now has new breeding values for individual i */
	    /* Cumulate squared Differences from old breeding values */
	    for (l = 0; l < NTraits; l++)
	    {
		z = b[l][i] - zz[l];
		d[l] += z * z;
		b[l][i] = zz[l];
	    }

	    n++;

	    /* Now correct tt[l] values for all individuals j, j<i, who */
            /* have off diagonal elements with i using the new breeding */
            /* values */

	    k = k1;
	    chsolve(sigch, zz, zz1, NTraits);
	    while (k > 0 && ginv[k].r1 == i)
	    {
		j = ginv[k].r2;
		z = ginv[k--].value;
		for (l = 0; l < NTraits; l++)
		    tt[l][j] -= z * zz1[l];
	    }
	}
	while (k > 0);

	/* Compute measure of distance between old and new breeding */
        /* value vectors...  */

	z = 0.0;
	for (l = 0; l < NTraits; l++)
	    z += d[l] / SigSq[l * (l + 1) / 2 + l];
	z /= (double) (n * NTraits);

	/* ...and check if converged yet */
	if (z <= Diff)
	    break;
    }
    free(tt);
    free(cc);
}

void breeding_values(double **z, double **a, double **b, double **e,
		                        int NTraits, Term * ginv, int nt)
{
    int i, j;

    itermo(z, b, NTraits, ginv, nt);

    for (j = 0; j < NTraits; ++j)
    {
	for (i = 0; i < pedsize; ++i)
	    a[j][qdex[i]] = b[j][i];
	for (i = 1; i <= np; ++i)
	    e[j][i] = z[j][i] - a[j][i];
    }
}

void mult_likval(int k, double log_det, double **z, double **a, double **b,
                                        Term * ginv, int nt, int NTraits)
{
    double lhood;
    double *inv, z1;
    int i, j, m, l;

    inv = malloc((NTraits + 1) * NTraits / 2 * sizeof(double));
    alloc_chk(WHERE, inv);

    lhood = NTraits * np * log(PPI);
    lhood += log_det;

    if (chfact(TauSq, inv, NTraits, &z1))
        give_up_where(WHERE, ERR_mxe, "mult_likval");

    chinv(inv, inv, NTraits);

    m = 0;
    for (i = 0; i < NTraits; ++i)
    {
	for (j = 0; j <= i; ++j)
	{
	    if (i == j)
		lhood += inv[m] * (innerprod(z[i], z[i], np) 
                                           - innerprod(a[j], a[j], np));
	    else
		lhood += 2.0 * inv[m] * (innerprod(z[i], z[j], np)
			        	   - innerprod(a[i], a[j], np));
	    ++m;
	}
    }
    if (chfact(SigSq, inv, NTraits, &z1))
        give_up_where(WHERE, ERR_mxg, "mult_likval");
    chinv(inv, inv, NTraits);

    m = 0;
    for (i = 0; i < NTraits; ++i)
    {
	for (j = 0; j <= i; ++j)
	{
	    if (i == j)
		lhood -= inv[m] * quarz(b[i], b[j], ginv, nt);
	    else
		lhood -= 2.0 * inv[m] * quarz(b[i], b[j], ginv, nt);
	    ++m;
	}
    }

    lhood = -0.5 * lhood;

    poly_print_pars(k, 5);
    printf("\n current log-likelihood =");
    print_val(lhood);
    printf("\n\n\n");
    fflush(stdout);

    free((void *) inv);
}

void em_iter(int k, double **a, double **b, double **epsi, double **z, 
             Term * ginv, int nt, double *evgen, int NTraits, 
             int *SigModel, int *TauModel)
{				/* EM iteration */
                                /* iterate for expected breeding values */
    int i, l;
    double temp;
    double log_det;
    double *ast, *tst;

    for (l = 0; l < NTraits; ++l)
	out_fixed(z[l], l, l);

    breeding_values(z, a, b, epsi, NTraits, ginv, nt);

    ast = malloc(NTraits * (NTraits + 1) / 2 * sizeof(double));
    alloc_chk(WHERE, ast);
    tst = malloc(NTraits * (NTraits + 1) / 2 * sizeof(double));
    alloc_chk(WHERE, tst);

    log_det = ss(evgen, ast, tst, NTraits);

    /* now evaluate likelihood */
    if ((k / ItWt) * ItWt == k)
    {
	mult_likval(k, log_det, z, a, b, ginv, nt, NTraits);
        if (ChkTrc)
        {
	    chk_sgfit(ginv, nt, b, ast, NTraits, SigModel);
            chk_tufit(epsi, tst, NTraits, TauModel);
        }
    }

    /* now implement EM equations for new variances  */
    sig_fit(ginv, nt, b, ast, NTraits, SigModel);

    tau_fit(epsi, tst, NTraits, TauModel);

    /* now backfit for each class of fixed effects */
    if (NFix == 0)
    {
	if (IPtEM == 1)
	{
	    for (l = 0; l < NTraits; ++l)
		out_fixed(z[l], l, l);
	    breeding_values(z, a, b, epsi, NTraits, ginv, nt);
	}
	for (l = 0; l < NTraits; ++l)
	    mu[l] = 0.0;
	for (l = 0; l < NTraits; ++l)
	    for (i = 1; i <= np; ++i)
	    {
		z[l][i] = (x[l][i] - a[l][i]);
		mu[l] += z[l][i];
	    }
	for (l = 0; l < NTraits; ++l)
	    mu[l] = mu[l] / np;
	for (l = 0; l < NTraits; ++l)
	    Mufti[l][0][1] = mu[l];
	if (IPtEM == 1)
	{
	    ss(evgen, ast, tst, NTraits);
	    tau_fit(epsi, tst, NTraits, TauModel);
	}
    } else
    {
	for (i = 1; i <= NFix; ++i)
	{
	    if (IPtEM == 1)
	    {
		for (l = 0; l < NTraits; ++l)
		    out_fixed(z[l], l, l);
		breeding_values(z, a, b, epsi, NTraits, ginv, nt);
	    }
	    for (l = 0; l < NTraits; ++l)
		out_bar_one(a[l], z[l], x[l], i, l);
	    for (l = 0; l < NTraits; ++l)
	    {
		temp = fixeffit(z[l], i, l);
		mu[l] = temp;
		Mufti[l][0][1] = temp;
	    }
	    if (IPtEM == 1)
	    {
		ss(evgen, ast, tst, NTraits);
		tau_fit(epsi, tst, NTraits, TauModel);
	    }
	}
    }
    free((void *) tst);
    free((void *) ast);
}

void setvals(double **b)
{
    int dd, l, i;

    for (l = 0; l < NTraits; ++l)
    {
	mu[l] = Mufti[l][0][1];
    }

    /* set initial breeding values */

    dd = 0;
    for (l = 0; l < NTraits; ++l)
    {
	for (i = 0; i < pedsize; ++i)
	    b[l][i] = (SigSq[dd] / (TauSq[dd] + SigSq[dd])) 
                                   * (x[l][qdex[i]] - mu[l]);
	dd += l + 2;
    }

    for (l = 0; l < NTraits; ++l)
	x[l][0] = 0.0;
}

/* sort the vector of eigenvalues */
void sortf(double *evgen1, double *evgen, int *n, int cpnp)
{
    int i, j, k;

    for (i = 1; i <= cpnp; ++i)
    {

	for (k = 1; k <= (*n); ++k)
	{
	    if (evgen1[i] < evgen[k])
		break;
	}

	for (j = (*n) + 1; j > k; --j)
	{
	    evgen[j] = evgen[j - 1];
	}

	evgen[k] = evgen1[i];
	++(*n);
    }
}

void nogene(int NTraits, double **a, double **z, int *TauModel)
{
    int l, i, j, m;
    double lhood, temp;
    double tauinv[MAX_TRAIT * (MAX_TRAIT + 1) / 2];


    /* use sample means as the estimates of fixed effects */
    if (NFix == 0)
    {
	for (l = 0; l < NTraits; ++l)
	    mu[l] = 0.0;
	for (l = 0; l < NTraits; ++l)
	    for (i = 1; i <= np; ++i)
	    {
		z[l][i] = (x[l][i] - a[l][i]);
		mu[l] += z[l][i];
	    }
	for (l = 0; l < NTraits; ++l)
	    mu[l] = mu[l] / np;
	for (l = 0; l < NTraits; ++l)
	    Mufti[l][0][1] = mu[l];
    } else
    {
	for (i = 1; i <= NFix; ++i)
	{
	    for (l = 0; l < NTraits; ++l)
		out_bar_one(a[l], z[l], x[l], i, l);
	    for (l = 0; l < NTraits; ++l)
	    {
		temp = fixeffit(z[l], i, l);
		mu[l] = temp;
		Mufti[l][0][1] = temp;
	    }
	}
    }
    /* take the residuals */
    for (l = 0; l < NTraits; ++l)
	out_fixed(z[l], l, l);

    /* use residual sample variance-covariance as estimates of residual */
    /* variance-cobvariance */

    m = 0;
    for (i = 0; i < NTraits; ++i)
	for (j = 0; j <= i; ++j)
	{
	    if (TauModel[m] == 0)
		TauSq[m] = 0.0;
	    else
		TauSq[m] = innerprod(z[i], z[j], np) / np;
	    SigSq[m] = 0.0;
	    m++;
	}

    /* compute the likelihood */

    (void) chfact(TauSq, tauinv, NTraits, &temp);
    lhood = -.5 * (NTraits * np * log(PPI) + np * temp + NTraits * np);

    printf("\n estimates of environmental model\n\n");

    poly_print_pars(0, 6);
    printf("\n environmental model log-likelihood =");
    print_val(lhood);
    printf("\n\n\n");
    fflush(stdout);

}

alloc_vect(double ***pa, double ***pepsi, double ***pz, double ***pb)
{
    int i;

    *pa = malloc(NTraits * sizeof(double *));
    alloc_chk(WHERE, *pa);
    *pepsi = malloc(NTraits * sizeof(double *));
    alloc_chk(WHERE, *pepsi);
    *pz = malloc(NTraits * sizeof(double *));
    alloc_chk(WHERE, *pz);
    *pb = malloc(NTraits * sizeof(double *));
    alloc_chk(WHERE, *pb);

    *(*pa) = malloc(NTraits * (np + 1) * sizeof(double));
    alloc_chk(WHERE, *(*pa));
    *(*pepsi) = malloc(NTraits * (np + 1) * sizeof(double));
    alloc_chk(WHERE, *(*pepsi));
    *(*pz) = malloc(NTraits * (np + 1) * sizeof(double));
    alloc_chk(WHERE, *(*pz));
    *(*pb) = malloc(NTraits * pedsize * sizeof(double));
    alloc_chk(WHERE, *(*pb));

    for (i = 1; i < NTraits; ++i)
    {
	*(*pa + i) = *(*pa + i - 1) + np + 1;
	*(*pepsi + i) = *(*pepsi + i - 1) + np + 1;
	*(*pz + i) = *(*pz + i - 1) + np + 1;
	*(*pb + i) = *(*pb + i - 1) + pedsize;
	**(*pz + i) = 0.0;
    }
}

free_vect(double **a, double **epsi, double **z, double **b,
                                     double *evgen, Term * ginverse)
{
    free((void *) *a);
    free((void *) a);
    free((void *) *epsi);
    free((void *) epsi);
    free((void *) *z);
    free((void *) z);
    free((void *) *b);
    free((void *) b);
    a = epsi = z = b = NULL;
    free((void *) evgen);
    evgen = NULL;
    free((void *) ginverse);
    ginverse = NULL;
}

main(int argc, char *argv[])
{
    int i, l, k, cpnp, *cdex, nevgen;
    int **dodo = NULL;
    double **a, **epsi, **z, **b;
    double *evgen, *evgen1;
    double log_det;
    int gsize;
    Ind **cpnghd;
    Term *ginverse;
    Indlist *peopl;

    atexit(print_tail);
    print_head(PROGNAM);
 
    /* process all of the input parameter statements */

    if (!proc_pars(argc, argv, PNAM))
	exit(EXIT_FAILURE);

    if (!proc_poly())
	exit(EXIT_FAILURE);

    if (!proc_fils())
	exit(EXIT_FAILURE);

    /* reads and checks the pedigree file for pedigree errors */

    if (!chk_pedfil())
        give_up(ERR_qit);	

    /* if the pedigree file was not in chronological order, */
    /* writes a new one */

    if (!mak_opedfil())
        give_up(ERR_qit);	

    /* reopen the pedigree file and read the pedigree data */

    if (!(pfile = open_ped()))
	give_up(ERR_qit);
                                 
    alloc_ped();
    input_data(pfile, 0, dodo);
    fclose(pfile);
    free_ped();

    /* the no. of observed individuals is known;  check against limit. 
     * also the no. of fixed effects levels is known;  check against no. 
     * starting values */

    if (!check_pars(np))
	exit(EXIT_FAILURE);

    /* Allocate space for eigenvalues of G-observed,  and the non-zero 
     * terms of G-full-inverse 
     *
     * evgen will now instead contain the eigenvalues of the INVERSE of 
     * G-obs;
     * logdet, sigtrc, and tautrc have been modified accordingly */

    ginverse = malloc(pedsize * 4 * sizeof(Term));
    alloc_chk(WHERE, ginverse);

    /* find non-zero terms of G-full-inverse for all pedigree individuals*/

    gsize = henderson(ginverse);
    ginverse = realloc(ginverse, (gsize + 1) * sizeof(Term));
    alloc_chk(WHERE, ginverse);

    /* For testing; print G-inverse, and if eigenvalues are not to be 
    /* printed, exit */

    if (ChkGnv)
    {
	printf(" G-inverse:\n\n");
	for (i = 1; i <= gsize; ++i)
	{
	    printf(" %5d %5d %5d", i, ginverse[i].r1, ginverse[i].r2);
	    print_val(ginverse[i].value);
	    printf("\n");
	}
	if (!ChkEig)
	    exit(EXIT_SUCCESS);
	else
	    printf("\n");
    }
    /* find or read eigenvalues of g-inverse for OBSERVED individuals  */

    if (ICh == 1)
    {
	           		/* cdex: component index */
	cdex = malloc(pedsize * sizeof(int));
	alloc_chk(WHERE, cdex);
				/* cpnghd: component nghd */
	cpnghd = malloc(max_cpsize * sizeof(Ind *));
	alloc_chk(WHERE, cpnghd);
				/* evgen1: eigenvaluess for  component*/
	evgen1 = malloc((max_cpsize + 1) * sizeof(double));
        alloc_chk(WHERE, evgen1);
				/* evgen: eigenvalues */
	evgen = malloc((np + 1) * sizeof(double));
	alloc_chk(WHERE, evgen);

	nevgen = 0;		/* number of evgen values in evgen[]*/

	for (k = 1; k < n_compnts + 1; k++)
	{
	    i = 0;
	    cpnp = 0;		/* number of observed in component */
	    peopl = cmplist[k].people;
	    while (peopl->next != NULL)
	    {
		peopl = peopl->next;
		cpnghd[i] = peopl->current;
		if (qdex[cpnghd[i]->index] != 0)
		    cpnp++;
		cdex[cpnghd[i]->index] = i;
		i++;
	    }

            if (i > cpnp)
	        bigcheig(evgen1, cpnp, i, cpnghd, cdex);
            else if (i == cpnp)
                give_warn(ERR_noo, i, cpnghd[0]->self);

	    sortf(evgen1, evgen, &nevgen, cpnp);
	}
	free((void *) evgen1);
	evgen1 = NULL;
	free((void *) cdex);
	cdex = NULL;
	free((void *) cpnghd);
	cpnghd = NULL;
	printf(" Eigenvalues have been computed\n\n\n");
    } else
    if (ICh == 0)
    {
	evgen = peigen;
	printf(" Eigenvalues have been read from parameter file\n\n\n");
    } else
    if (ICh == -1)
    {
	if (!input_eigen(np))
	    exit(EXIT_FAILURE);
	else
	{
	    evgen = peigen;
	    printf(" Eigenvalues have been read from eigenvalues "
                                                          "file\n\n\n");
	}
    }

    free_cmplists();
    free_nghds();

    /* For testing, print terms and exit	 */
    if (ChkEig)
    {
	printf(" eigenvalues of G-observed:\n\n");
	for (i = 1; i <= np; ++i)
	{
	    print_val5(evgen[i]);
	    if (i % 10 == 0)
		printf("\n");
	}
	printf("\n\n");
	exit(EXIT_SUCCESS);
    }
    for (i = 1; i <= np; ++i)
	evgen[i] = 1.0 / evgen[i];

    /* allocate space for vectors */

    alloc_vect(&a, &epsi, &z, &b);

    /* set initial values */

    setvals(b);

    /* full or partitioned EM ? */

    if (IPtEM == 1)
        printf(" using partitioned EM\n\n");
    else
        printf(" using full EM\n\n");

    /* now ready for EM iteration  */

    for (k = 0; k <= ItEM; ++k)
    {
	em_iter(k, a, b, epsi, z, ginverse, gsize, evgen, NTraits, 
                                                   SigModel, TauModel);
    }

    for (l = 0; l < NTraits; ++l) 
        out_fixed(z[l], l, l); 
    breeding_values(z, a, b, epsi, NTraits, ginverse, gsize);
    log_det = ss(evgen, 0, 0, NTraits); 
    mult_likval(k, log_det, z, a, b, ginverse, gsize, NTraits);

    /* fit an environmental model if IEn = 1 */

    if (IEn == 1)
    {
	for (l = 0; l < NTraits; ++l)
	    for (i = 1; i <= np; ++i)
		a[l][i] = 0.0;
	nogene(NTraits, a, z, TauModel);
    }
      
    free_input_data();
    free_vect(a, epsi, z, b, evgen, ginverse);

    return 0;
}
 
